/*****************************************************************************************
 *
 *  Copyright (C) 2003/2004 MADos Core Developer Team
 *  Copyright 2006+ by Cieszak [NokiX version]
 *  Modified and optimized by NuukiaWorld
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ****************************************************************************************/

#include "game.h"
#include "menu.h"

// this global variable keeps track of the game, even if you quit the whole app
struct game *game;

//all shapes definitions
typedef struct shape_table
{
   unsigned char table[4][4];    // these fit in char, save flash
   unsigned char width;          // bitfields removed
   unsigned char height;         // structure is padded to anyway to multiple of 4 --> 24 bytes!
   unsigned char pointv[4];
} shape_t;
static const shape_t shape[10] = {   // tetris_define_shapes() and dynamic RAM replaced with const
   { { { 0, 4, 0, 4 }, {15, 4,15, 4 }, { 0, 4, 0, 4 }, { 0, 4, 0, 4 } }, 4, 1, { 5, 8, 5, 8 } },
   { { {12,12,12,12 }, {12,12,12,12 }, { 0, 0, 0, 0 }, { 0, 0, 0, 0 } }, 2, 2, { 6, 6, 6, 6 } },
   { { { 4, 4, 0, 4 }, {14, 6,14,12 }, { 0, 4, 4, 4 }, { 0, 0, 0, 0 } }, 3, 2, { 5, 5, 6, 5 } },
   { { {12, 4,12, 4 }, { 6,12, 6,12 }, { 0, 8, 0, 8 }, { 0, 0, 0, 0 } }, 3, 2, { 6, 7, 6, 7 } },
   { { { 6, 8, 6, 8 }, {12,12,12,12 }, { 0, 4, 0, 4 }, { 0, 0, 0, 0 } }, 3, 2, { 6, 7, 6, 7 } },
   { { { 2, 8,14,12 }, {14, 8, 8, 4 }, { 0,12, 0, 4 }, { 0, 0, 0, 0 } }, 3, 2, { 6, 7, 6, 7 } },
   { { {14, 4, 8,12 }, { 2, 4,14, 8 }, { 0,12, 0, 8 }, { 0, 0, 0, 0 } }, 3, 2, { 6, 7, 6, 7 } },
   { { { 8, 8, 8, 8 }, { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, { 0, 0, 0, 0 } }, 1, 1, { 2, 2, 2, 2 } },
   { { {12, 4, 8,12 }, { 4,12,12, 8 }, { 0, 0, 0, 0 }, { 0, 0, 0, 0 } }, 2, 2, { 4, 3, 3, 4 } },
   { { {14, 4,14, 4 }, { 0, 4, 0, 4 }, { 0, 4, 0, 4 }, { 0, 0, 0, 0 } }, 3, 1, { 5, 5, 5, 5 } }
};

#define OMSG_TIMER   (0x1f00)
static const ULONG timer=OMSG_TIMER;
static const ULONG key_timer=MSG_KEYPRESS;

extern const ringtone_t tone_drop;
extern const ringtone_t tone_line;

static const int beginlev = 0;
static const char timings[] = { 6, 6, 6, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2};
static const char blockbmp[3][9] = { {1, 1, 1, 1, 0, 1, 1, 1, 1}, {1, 0, 1, 0, 1, 1, 1, 1, 1}, {1, 1, 1, 1, 1, 1, 1, 1, 1} };
static const char blockbmp96[3][16] = { {1, 1, 1, 1,  1, 0, 0, 1,  1, 0, 0, 1,  1, 1, 1, 1}, {1, 0, 1, 1,  0, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1}, {1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1} };

// All local functions made "static" to allow compiler to optimize the calls
static void tetris_init_game ( void );
static void tetris_create_shape ( void );
static void tetris_show_score ( void );
static void tetris_show_next ( void );
static void tetris_print_shape ( int shape_no, int x, int y, int rot, int col );
static void tetris_set1pixel ( int x, int y, int col );
static void tetris_left_proc ( void );
static void tetris_right_proc ( void );
static void tetris_anti_proc ( void );
static void tetris_fast_proc ( void );
static int tetris_block_can_drop ( int shape_no, int xpos, int ypos, int rot );
static int tetris_block_can_left ( int shape_no, int xpos, int ypos, int rot );
static int tetris_block_can_right ( int shape_no, int xpos, int ypos, int rot );
static void tetris_remove_full_lines ( int y );
static int tetris_check_rot ( int shape_no, int xpos, int ypos, int newrot );
static void tetris_print_all_blocks ( void );
static void tetris_drop_block ( void );
static void tetris_end_game ( void );
static void tetris_create_shape ( void );
static void tetris_store_shape ( int shape_no, int xpos, int ypos, int rot );
static int tetris_level ( void );
static void tetris_pause(int pause);
static void tetris_print_board(void);
static void tetris_print_stats(void);
static void tetris_print_hint(int x, int t);

static void tetris_start_timer(void);
static void tetris_stop_timer(void);

/**************************************************************************************/

/* frees the game structure */
void free_game(void)
{
   if(game)
   {
      free(game);
      game=NULL;
//      free(shape);	   // no need to free shape anymore
//      shape=NULL;		// also, usually it wasn't freed at all!
   }
}

/**************************************************************************************/

/* UI messages dispatcher */
extern rexx_uint game_dispatcher_id; /* dispatcher's call id */
rexx("game_dispatcher_id=add_5e0_id(game_dispatcher,locate('action_back_back'))"); /* auto creation of the id */

int game_dispatcher(int msg)
{
   //printf("*GAME DISPATCHER\n");
   switch(msg)
   {
      case MSG_D_INIT: // initialize game
         freegfx_open();

         app.old_lights_owner=lights_owner;
         lights_owner=LIGHTS_OWNER_ON;
         if(own_get_game_setting(GS_LIGHTS)) send_message(MSG_LIGHTS_ON);
         else send_message(MSG_LIGHTS_OFF);
         
         if(!game->initialized)
         {
//            tetris_define_shapes();	// shapes[] is now const
            tetris_init_game();
            tetris_pause(1);
            tetris_pause(0);
            game->initialized=1;
         }
         else
         {
            //redraw screen
            tetris_show_next();
            tetris_show_score();
            tetris_print_board();
            tetris_print_all_blocks();
            //powrot do gry, pause
            tetris_pause(0);
            tetris_pause(1);
         }

         msg=MSG_R_STAY;
         break;

      case MSG_D_QUIT: // handle quit to menu
         lights_owner=app.old_lights_owner;
         send_message(MSG_TIMED_LIGHTS_ON);

         freegfx_close();
         tetris_pause(1);

         msg=MSG_R_QUIT;
         break;

      case MSG_DISPLAY_MESSAGE: // quits to menu when message like "Charging" is displayed
         tetris_pause(1);
         send_message(MSG_REQUEST_QUIT);
         break;

      case MSG_INCOMING_CALL: // quits to menu when someone calls
         tetris_pause(1);
         lights_owner=app.old_lights_owner;
         send_message(MSG_TIMED_LIGHTS_ON);
   
         freegfx_close();
         break;

      case MSG_CALL_FINISHED: // comes after finishing the call
         msg=MSG_R_LEAVE;
         break;
         
      case OMSG_TIMER:
//         printf("OMSG_TIMER\n");
         game->dlyloop++;
         tetris_start_timer();
         if ( game->dlyloop >= timings[tetris_level (  )] )
         {
            game->dlyloop = 0;
            tetris_drop_block (  );
         }
         break;
         
      case MSG_KEYRELEASE:
         own_timer_abort(&key_timer);
         game->key=0xFF;
         break;

      case MSG_KEYPRESS:
         if (game->paused)
            tetris_pause(0);
         else
         if(game->gameover)
         {
            //stopping timer used for show "record" message
//            own_timer_abort(&tim_disp_msg);				// Record showing moved to menu.c
//            send_message(MSG_CLOSE_DISPLAY_MESSAGE);

//            if(own_get_unify_keycode(UKM_NPM5_CURSORS1)==KEY_NAVI)
//               send_message(MSG_D_QUIT);
         }
         else
         if(!game->lock_keys)
         {
            if (game->key==0xFF) //user press new key
            {
               game->key=own_get_unify_keycode(UKM_NPM5_CURSORS1);
               game->keycounter=0;
            }
            if((game->key==KEY_1)||(game->key==KEY_4)||(game->key==KEY_3)||(game->key==KEY_6))
            {
               if (game->keycounter==0)
                  own_timer_start(&key_timer,TT_MESSAGE,30);
               else
                  own_timer_start(&key_timer,TT_MESSAGE,8);
               game->keycounter++;
            }

            switch(game->key) // also maps 5510 cursors (x,shift,chr,*) as keys 2,4,8,6
            {
               case KEY_1:
               case KEY_4:
                  tetris_left_proc();
                  break;

               case KEY_3:
               case KEY_6:
                  tetris_right_proc (  );
                  break;

               case KEY_5:
               case KEY_8:
               case KEY_0:
                  tetris_fast_proc (  );
                  break;

               case KEY_2:
               case KEY_DOWN:
                  tetris_anti_proc (  );
                  break;
            }
         }
         break;

      case 0x1B59:	// suppress keypad tones
      case MSG_KEYPRESS_CHAR:
         msg=MSG_NOP; // prevents our dispatcher from quiting when numeric keys are being pressed
         break;
   }

   return msg;
}

// void tetris_define_shapes ( void )	// removed, shapes[] is now const, no need to initialize!

static void tetris_init_game ( void )
{
   #if defined DEBUG
      printf("tetris_init_game\n");
   #endif

   srand(get_timestamp());
   int i, j;

   //score_position = -1;
   game->dlyloop=game->gameover=game->paused=game->rows = game->score = game->shape_no = game->rot = game->xpos = game->ypos = 0;
   for ( i = 0; i < UWIDTH; i++ )
      for ( j = 0; j < UHEIGHT; j++ )
         game->grid[i][j] = 0;
   for ( i = 0; i < 10; i++ )
         game->stats[i]=0;
   tetris_create_shape (  );   /* Set up 1st shape */
   tetris_create_shape (  );   /* Set up next shape */

   tetris_print_stats();
   tetris_print_board();
   tetris_show_score();
   tetris_show_next();
}

static void tetris_create_shape ( void )
{
   #if defined DEBUG
      printf("tetris_create_shape\n");
   #endif

    game->shape_no = game->next_no;
    game->rot = game->next_rot;
    game->next_no = rand()%(settings->total_shapes);
    game->stats[game->next_no]++;
    game->next_rot = rand()%4;
    game->xpos = ( UWIDTH >> 1 ) - 1;
    game->ypos = -1;
    tetris_print_stats();
   #if defined DEBUG
      printf("created shape: no:%d rot:%d\n", game->next_no, game->next_rot);
   #endif

}

static void tetris_show_score ( void )
{
   char score_buf[6], level_buf[3], rows_buf[4];

   freegfx_begin(FALSE);
   own_draw_box(13*HUNIT+NORM_OFFSET, VUNIT+NEXT_HEIGHT+6+0*8, 5*5, 1*8+1, DM_CLEAR);
   own_draw_box(13*HUNIT+NORM_OFFSET, VUNIT+NEXT_HEIGHT+6+1*8+1, 5*5, 2*8+1, DM_CLEAR);
   own_draw_box(13*HUNIT+NORM_OFFSET, VUNIT+NEXT_HEIGHT+6+2*8+2, 5*5, 3*8+1, DM_CLEAR);
   
   if (game->score == 0 )
   {
      own_print_string((const nokstr_t*)"score", 13*HUNIT+NORM_OFFSET, VUNIT+NEXT_HEIGHT+6+0*8+6, app.font_plain, DM_SET);	// font ptrs are stored in app for quicker access
      own_print_string((const nokstr_t*)"level", 13*HUNIT+NORM_OFFSET, VUNIT+NEXT_HEIGHT+6+1*8+6+1, app.font_plain, DM_SET);
      own_print_string((const nokstr_t*)"rows", 13*HUNIT+NORM_OFFSET, VUNIT+NEXT_HEIGHT+6+2*8+6+2, app.font_plain, DM_SET);
   }
   else
   {
      ntoa(game->score, score_buf);
      ntoa(beginlev+1+(game->rows/10), level_buf);
      ntoa(game->rows, rows_buf);

      own_print_string((const nokstr_t*)"0000", 13*HUNIT+NORM_OFFSET, VUNIT+NEXT_HEIGHT+6+0*8+6, app.font_plain, DM_SET);
      own_print_string((const nokstr_t*)"00", 13*HUNIT+NORM_OFFSET+10, VUNIT+NEXT_HEIGHT+6+1*8+6+1, app.font_plain, DM_SET);
      own_print_string((const nokstr_t*)"000", 13*HUNIT+NORM_OFFSET+5, VUNIT+NEXT_HEIGHT+6+2*8+6+2, app.font_plain, DM_SET);

      own_print_string((const nokstr_t*)score_buf, 13*HUNIT+NORM_OFFSET+15-(strlen(score_buf)*5)+5, VUNIT+NEXT_HEIGHT+6+0*8+6, app.font_plain, DM_SET);
      own_print_string((const nokstr_t*)level_buf, 13*HUNIT+NORM_OFFSET+15-(strlen(level_buf)*5)+5, VUNIT+NEXT_HEIGHT+6+1*8+6+1, app.font_plain, DM_SET);
      own_print_string((const nokstr_t*)rows_buf, 13*HUNIT+NORM_OFFSET+15-(strlen(rows_buf)*5)+5, VUNIT+NEXT_HEIGHT+6+2*8+6+2, app.font_plain, DM_SET);
   }
   freegfx_end(false);
}

static void tetris_show_next ( void )
{
   #if defined DEBUG
      printf("tetris_show_next\n");
   #endif

   freegfx_begin(FALSE);
   own_draw_box(13*HUNIT+NORM_OFFSET-2, VUNIT-1, NEXT_HEIGHT+1, VUNIT+NEXT_HEIGHT+1, DM_CLEAR);
   own_draw_rect(13*HUNIT+NORM_OFFSET-2, VUNIT-1, 13*HUNIT+NORM_OFFSET-2+NEXT_HEIGHT+1, VUNIT+NEXT_HEIGHT+1,1, DM_SET);
   freegfx_end(false);
   
   tetris_print_shape(game->next_no, 13, 1,game->next_rot, 1);
}

static void tetris_print_shape ( int shape_no, int x, int y, int rot, int col )
{
   #if defined DEBUG
      printf("tetris_print_shape\n");
   #endif

    int x1, x2, x3, x4, y1;
    int t0, t1, t2, t3;

    t0 = shape[shape_no].table[0][rot]; /* Bit map of 1st Row */
    t1 = shape[shape_no].table[1][rot]; /* Bit map of 2nd Row */
    t2 = shape[shape_no].table[2][rot]; /* Bit map of 3rd Row */
    t3 = shape[shape_no].table[3][rot]; /* Bit map of 4th Row */

    x1 = x;                     /* Position of 1st column of block grid */
    x2 = x1 + 1;                /* Position of 2nd column of block grid */
    x3 = x2 + 1;                /* Position of 3rd column of block grid */
    x4 = x3 + 1;                /* Position of 4th column of block grid */
    y1 = y;                     /* Position of 1st row of block grid    */

    if ( y > -1 )
    {
        if ( t0 & 8 )
            tetris_set1pixel ( x1, y1, col );
        if ( t0 & 4 )
            tetris_set1pixel ( x2, y1, col );
        if ( t0 & 2 )
            tetris_set1pixel ( x3, y1, col );
        if ( t0 & 1 )
            tetris_set1pixel ( x4, y1, col );
    }
    y1 += 1;                    /* Position of next row */
    if ( y > -2 )
    {
        if ( t1 & 8 )
            tetris_set1pixel ( x1, y1, col );
        if ( t1 & 4 )
            tetris_set1pixel ( x2, y1, col );
        if ( t1 & 2 )
            tetris_set1pixel ( x3, y1, col );
        if ( t1 & 1 )
            tetris_set1pixel ( x4, y1, col );
    }
    y1 += 1;                    /* Position of next row */
    if ( y > -3 )
    {
        if ( t2 & 8 )
            tetris_set1pixel ( x1, y1, col );
        if ( t2 & 4 )
            tetris_set1pixel ( x2, y1, col );
        if ( t2 & 2 )
            tetris_set1pixel ( x3, y1, col );
        if ( t2 & 1 )
            tetris_set1pixel ( x4, y1, col );
    }
    y1 += 1;                    /* Position of next row */
    if ( y > -4 )
    {
        if ( t3 & 8 )
            tetris_set1pixel ( x1, y1, col );
        if ( t3 & 4 )
            tetris_set1pixel ( x2, y1, col );
        if ( t3 & 2 )
            tetris_set1pixel ( x3, y1, col );
        if ( t3 & 1 )
            tetris_set1pixel ( x4, y1, col );
    }
    
   //show hint at bottom of board
   if((col==1) && (settings->hint))
      tetris_print_hint(x, t0|t1|t2|t3);
}


/****  direct screen functions  ***/

static void tetris_set1pixel ( int x, int y, int col )
{
   freegfx_begin(false);
   int i, j;
   if ( col == 1 )
   {
      if(S96)
      {
         for ( i = 0; i < 4; i++ )
            for ( j = 0; j < 4; j++ )
               draw_pixel(x*HUNIT+j, y*VUNIT+i, blockbmp96[settings->blockdesign][4 * i + j]);
      }
      else
      {
         for ( i = 0; i < 3; i++ )
            for ( j = 0; j < 3; j++ )
               draw_pixel(x*HUNIT+j, y*VUNIT+i, blockbmp[settings->blockdesign][3 * i + j]);
      }
   }
   else
      own_draw_box(x*HUNIT, y*VUNIT, HUNIT, VUNIT, DM_CLEAR);
   freegfx_end(false);
}

static void tetris_print_hint (int x, int t)
{
   if((x>=0)&&(x<=UWIDTH))
   {
      freegfx_begin(false);
      //clear hint
      own_draw_box(1*HUNIT, VUNIT*UHEIGHT+1, UWIDTH*VUNIT+1, 1+(S96), DM_CLEAR);
      //draw hint
      if(t&8)
         own_draw_box(x*HUNIT, VUNIT*UHEIGHT+1, 3+(S96), 1+(S96), DM_SET);
      x++;
      if (t&4)
         own_draw_box(x*HUNIT, VUNIT*UHEIGHT+1, 3+(S96), 1+(S96), DM_SET);
      x++;
      if (t&2)
         own_draw_box(x*HUNIT, VUNIT*UHEIGHT+1, 3+(S96), 1+(S96), DM_SET);
      x++;
      if (t&1)
         own_draw_box(x*HUNIT, VUNIT*UHEIGHT+1, 3+(S96), 1+(S96), DM_SET);
      freegfx_end(false);
   }
}

static void tetris_print_all_blocks ( void )
{
   #if defined DEBUG
      printf("tetris_print_all_blocks\n");
   #endif

   int i, j;
   for ( i = 0; i < UHEIGHT; i++ )
      for ( j = 0; j < UWIDTH; j++ )
         if ( game->grid[j][i] )
            tetris_set1pixel ( j + NORM_OFFSET, i, 1 );
         else
            tetris_set1pixel ( j + NORM_OFFSET, i, 0 );
}

static void tetris_print_board(void)
{
   freegfx_begin(false);
   own_draw_rect(0, 0, (UWIDTH+2)*HUNIT-1, (UHEIGHT+1)*VUNIT-1, 1, DM_SET);
   own_draw_rect(HUNIT-1, 0, (UWIDTH+1)*HUNIT, UHEIGHT*VUNIT, 1, DM_SET);
   own_draw_rect(HUNIT, 0, (UWIDTH+1)*HUNIT-1, UHEIGHT*VUNIT-1, 1, DM_CLEAR);
   freegfx_end(false);
}

static void tetris_print_stats(void)
{
   if(settings->stats)
   {
      //find max value for auto scroll
      int i;
      int min=game->stats[0];
      for(i=0; i<settings->total_shapes; i++)
         if(min>game->stats[i])
            min=game->stats[i];

      if(min>40)
      {
         for(i=0; i<settings->total_shapes; i++)
            game->stats[i]-=20;
      }

      freegfx_begin(false);
      own_draw_box(23*HUNIT+NORM_OFFSET,0, 10, screen_info.height, DM_CLEAR);
      for(i=0; i<settings->total_shapes; i++)
         own_draw_line(23*HUNIT+NORM_OFFSET-1+i, 0, 23*HUNIT+NORM_OFFSET-1+i, game->stats[i], 1, DM_SET);
      freegfx_end(false);
   }
}


/****  calculating functions  ***/

static void tetris_left_proc ( void )
{
//   printf("tetris_left_proc\n");
    if ( tetris_block_can_left ( game->shape_no, game->xpos, game->ypos, game->rot ) )
    {
        tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 0 );
        game->xpos--;
        tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 1 );
    }
}

static void tetris_right_proc ( void )
{
//   printf("tetris_right_proc\n");
       if ( tetris_block_can_right ( game->shape_no, game->xpos, game->ypos, game->rot ) )
    {
        tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 0 );
        game->xpos++;
        tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 1 );
    }
}

static void tetris_anti_proc ( void )
{
   int newrot;

   newrot = (game->rot+3)%4;
   if ( tetris_check_rot ( game->shape_no, game->xpos, game->ypos, newrot ) )
   {
      tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 0 );
      game->rot = newrot;
      tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 1 );
   }
   else
   //checking wheter we can rotate if we move shape left or right
   //right
   if ( tetris_check_rot ( game->shape_no, game->xpos+1, game->ypos, newrot ) )
   {
      tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 0 );
      game->xpos++;
      game->rot = newrot;
      tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 1 );
   }
   else
   //left
   if ( tetris_check_rot ( game->shape_no, game->xpos-1, game->ypos, newrot ) )
   {
      tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 0 );
      game->xpos--;
      game->rot = newrot;
      tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 1 );
   }
}


static void tetris_fast_proc ( void )
{
   //lock keyboard, user cannot move or rotate shape
   //keys will be unclocked in tetris_drop_block
   game->lock_keys=1;

   while(tetris_block_can_drop(game->shape_no, game->xpos, game->ypos, game->rot) /*&& kpd_getkey (  ) != 0xFF */)
   {
      tetris_print_shape(game->shape_no, game->xpos+NORM_OFFSET, game->ypos, game->rot, 0);
      game->ypos++;
      tetris_print_shape(game->shape_no, game->xpos+NORM_OFFSET, game->ypos, game->rot, 1);
/*        usleep ( 50000 ); */
      int i;
      for(i=0; i<500; i++);
   }
   //don't wait for timer to saveblock, remove lines and redraw screen)
   tetris_drop_block();
}

static int tetris_block_can_drop ( int shape_no, int xpos, int ypos, int rot )
{
//   printf("tetris_block_can_drop\n");
       int y1, c;

    c = 3;

    while ( ( c >= 0 ) && ( ( shape[shape_no].table[c][rot] & 8 ) == 0 ) )
        c--;
    y1 = ypos + c + 1;
    if ( ( c != -1 ) && ( y1 >= 0 ) )
        if ( ( y1 == UHEIGHT ) || ( game->grid[xpos][y1] != 0 ) )
            return ( 0 );

    c = 3;
    while ( ( c >= 0 ) && ( ( shape[shape_no].table[c][rot] & 4 ) == 0 ) )
        c--;
    y1 = ypos + c + 1;
    if ( ( c != -1 ) && ( y1 >= 0 ) )
        if ( ( y1 == UHEIGHT ) || ( game->grid[xpos + 1][y1] != 0 ) )
            return ( 0 );

    c = 3;
    while ( ( c >= 0 ) && ( ( shape[shape_no].table[c][rot] & 2 ) == 0 ) )
        c--;
    y1 = ypos + c + 1;
    if ( ( c != -1 ) && ( y1 >= 0 ) )
        if ( ( y1 == UHEIGHT ) || ( game->grid[xpos + 2][y1] != 0 ) )
            return ( 0 );

    c = 3;
    while ( ( c >= 0 ) && ( ( shape[shape_no].table[c][rot] & 1 ) == 0 ) )
        c--;
    y1 = ypos + c + 1;
    if ( ( c != -1 ) && ( y1 >= 0 ) )
        if ( ( y1 == UHEIGHT ) || ( game->grid[xpos + 3][y1] != 0 ) )
            return ( 0 );

    return 1;
}

static int tetris_block_can_left ( int shape_no, int xpos, int ypos, int rot )
{
//   printf("tetris_block_can_left\n");
    int x1, c;
    int y0, y1, y2, y3;
    int t0, t1, t2, t3;

    t0 = shape[shape_no].table[0][rot]; /* Bit map of 1st Row */
    t1 = shape[shape_no].table[1][rot]; /* Bit map of 2nd Row */
    t2 = shape[shape_no].table[2][rot]; /* Bit map of 3rd Row */
    t3 = shape[shape_no].table[3][rot]; /* Bit map of 4th Row */

    y0 = ypos;
    y1 = ypos + 1;
    y2 = ypos + 2;
    y3 = ypos + 3;

    c = 3;
    while ( ( c >= 0 ) && ( ( t0 & ( 1 << c ) ) == 0 ) )
        c--;
    x1 = xpos - 1 + ( 3 - c );
    if ( c != -1 )
        if ( ( x1 < 0 ) || ( ( y0 >= 0 ) && ( game->grid[x1][y0] != 0 ) ) )
            return ( 0 );

    c = 3;
    while ( ( c >= 0 ) && ( ( t1 & ( 1 << c ) ) == 0 ) )
        c--;
    x1 = xpos - 1 + ( 3 - c );
    if ( c != -1 )
        if ( ( x1 < 0 ) || ( ( y1 >= 0 ) && ( game->grid[x1][y1] != 0 ) ) )
            return ( 0 );

    c = 3;
    while ( ( c >= 0 ) && ( ( t2 & ( 1 << c ) ) == 0 ) )
        c--;
    x1 = xpos - 1 + ( 3 - c );
    if ( c != -1 )
        if ( ( x1 < 0 ) || ( ( y2 >= 0 ) && ( game->grid[x1][y2] != 0 ) ) )
            return ( 0 );

    c = 3;
    while ( ( c >= 0 ) && ( ( t3 & ( 1 << c ) ) == 0 ) )
        c--;
    x1 = xpos - 1 + ( 3 - c );
    if ( c != -1 )
        if ( ( x1 < 0 ) || ( ( y3 >= 0 ) && ( game->grid[x1][y3] != 0 ) ) )
            return ( 0 );

    return 1;
}

static int tetris_block_can_right ( int shape_no, int xpos, int ypos, int rot )
{
//   printf("tetris_block_can_right\n");
    int x1, c;
    int y0, y1, y2, y3;
    int t0, t1, t2, t3;

    t0 = shape[shape_no].table[0][rot]; /* Bit map of 1st Row */
    t1 = shape[shape_no].table[1][rot]; /* Bit map of 2nd Row */
    t2 = shape[shape_no].table[2][rot]; /* Bit map of 3rd Row */
    t3 = shape[shape_no].table[3][rot]; /* Bit map of 4th Row */

    y0 = ypos;
    y1 = ypos + 1;
    y2 = ypos + 2;
    y3 = ypos + 3;

    c = 0;
    while ( ( c < 4 ) && ( ( t0 & ( 1 << c ) ) == 0 ) )
        c++;
    x1 = xpos + 1 + ( 3 - c );
    if ( ( c != 4 ) && ( x1 >= 0 ) )
        if ( ( x1 == UWIDTH ) || ( ( y0 >= 0 ) && ( game->grid[x1][y0] != 0 ) ) )
            return ( 0 );

    c = 0;
    while ( ( c < 4 ) && ( ( t1 & ( 1 << c ) ) == 0 ) )
        c++;
    x1 = xpos + 1 + ( 3 - c );
    if ( ( c != 4 ) && ( x1 >= 0 ) )
        if ( ( x1 == UWIDTH ) || ( ( y1 >= 0 ) && ( game->grid[x1][y1] != 0 ) ) )
            return ( 0 );

    c = 0;
    while ( ( c < 4 ) && ( ( t2 & ( 1 << c ) ) == 0 ) )
        c++;
    x1 = xpos + 1 + ( 3 - c );
    if ( ( c != 4 ) && ( x1 >= 0 ) )
        if ( ( x1 == UWIDTH ) || ( ( y2 >= 0 ) && ( game->grid[x1][y2] != 0 ) ) )
            return ( 0 );

    c = 0;
    while ( ( c < 4 ) && ( ( t3 & ( 1 << c ) ) == 0 ) )
        c++;
    x1 = xpos + 1 + ( 3 - c );
    if ( ( c != 4 ) && ( x1 >= 0 ) )
        if ( ( x1 == UWIDTH ) || ( ( y3 >= 0 ) && ( game->grid[x1][y3] != 0 ) ) )
            return ( 0 );

    return 1;
}

static void tetris_remove_full_lines ( int y )
{
//   printf("tetris_remove_full_lines\n");
    int y1, y2, x;

    for ( y1 = y; y1 < y + 4 && y1 < UHEIGHT; y1++ )
    {
        for ( x = 0; x < UWIDTH; x++ )
            if ( game->grid[x][y1] == 0 )
                break;
        if ( x == UWIDTH )
        {
            for ( y2 = y1; y2 > 0; y2-- )
                for ( x = 0; x < UWIDTH; x++ )
                    game->grid[x][y2] = game->grid[x][y2 - 1];
            for ( x = 0; x < UWIDTH; x++ )
                game->grid[x][0] = 0;
            tetris_print_all_blocks (  );
            game->rows++;
        }
    }
}

static int tetris_check_rot ( int shape_no, int xpos, int ypos, int newrot )
{
//   printf("tetris_check_rot\n");
    int i;
    int ti;                     /* Bit map of i'th row    */
    int yi;                     /* Y position on i'th row */
    int x0, x1, x2, x3;

    x0 = xpos;
    x1 = xpos + 1;
    x2 = xpos + 2;
    x3 = xpos + 3;
    yi = ypos;

    for ( i = 0; i < 4; yi++, i++ )
    {
        if ( ( yi ) >= 0 )
        {
            ti = shape[shape_no].table[i][newrot];
            if ( ti & 8 )
                if ( ( x0 < 0 ) || ( x0 >= UWIDTH ) || ( game->grid[x0][yi] == 1 ) )
                    return 0;
            if ( ti & 4 )
                if ( ( x1 < 0 ) || ( x1 >= UWIDTH ) || ( game->grid[x1][yi] == 1 ) )
                    return 0;
            if ( ti & 2 )
                if ( ( x2 < 0 ) || ( x2 >= UWIDTH ) || ( game->grid[x2][yi] == 1 ) )
                    return 0;
            if ( ti & 1 )
                if ( ( x3 < 0 ) || ( x3 >= UWIDTH ) || ( game->grid[x3][yi] == 1 ) )
                    return 0;
        }
    }
    return 1;
}


static void tetris_drop_block ( void )
{
   //printf("tetris_drop_block\n");
    int orows;

    if ( tetris_block_can_drop ( game->shape_no, game->xpos, game->ypos, game->rot ) )
        tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos++, game->rot, 0 );
    else
    {
       //unlock keys if are locked by tetris_fast_proc
       game->lock_keys=0;

        if ( game->ypos < 0 )
            tetris_end_game (  );
        else
        {
           play_sound(&tone_drop);
            orows = game->rows;
            game->score += (shape[game->shape_no].pointv[game->rot])*(settings->level);
            tetris_store_shape ( game->shape_no, game->xpos, game->ypos, game->rot );
            tetris_remove_full_lines ( game->ypos );
            tetris_create_shape (  );
            tetris_show_score (  );
            tetris_show_next (  );
            if ( orows != game->rows )
               play_sound(&tone_line);
        }
    }
    tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 1 );
}


static void tetris_end_game ( void )
{
   //printf("tetris_end_game\n");
   tetris_stop_timer();
   game->gameover = 1;
	// top score message handling moved to menu.c
	// it is matter of taste but now the message does not appear immediately when the game is over
   freegfx_begin(false);
   own_draw_box(4, 3, 75, 11, DM_CLEAR);
   own_draw_rect(4, 3, 80, 14, 1, DM_SET);
   own_print_string((const nokstr_t*)"Game Over!", 13, 5+6, app.font_bold, DM_SET);
   play_sound(&tone_game_lost);
   freegfx_end(false);
}

static void tetris_store_shape ( int shape_no, int xpos, int ypos, int rot )
{
   int i;
   for ( i = 0; i < 4; i++ )
   {
      if ( shape[shape_no].table[i][rot] & 8 )
         game->grid[xpos][ypos + i] = 1;
      if ( shape[shape_no].table[i][rot] & 4 )
         game->grid[xpos + 1][ypos + i] = 1;
      if ( shape[shape_no].table[i][rot] & 2 )
         game->grid[xpos + 2][ypos + i] = 1;
      if ( shape[shape_no].table[i][rot] & 1 )
         game->grid[xpos + 3][ypos + i] = 1;
   }
}

static int tetris_level ( void )
{
   int level;
   level = game->rows/10+ beginlev;
   if ( level >= sizeof ( timings ) )
      level = sizeof ( timings ) - 1;
   return level;
}

static void tetris_pause(int pause)
{
   if(game)
   {
      //if paused->unpause; if unpaused->pause
      if (pause==-1)
         pause=(game->paused?0:1);
         
      if (pause!=game->paused)
      {
         if(game->paused)
         {
            game->paused=0;
            tetris_start_timer();
            
            //hide "pause" text
            freegfx_begin(false);
            own_draw_box(13*HUNIT+NORM_OFFSET, 19, 32 , 11, DM_CLEAR);
            freegfx_end(false);
            
            tetris_show_next();
            tetris_show_score();
            tetris_print_stats();
            tetris_print_board();
            tetris_print_all_blocks();
            #if defined DEBUG
               printf("Game unpaused\n");
            #endif
         }
         else
         {
            tetris_stop_timer();
            game->paused=1;

            //show "pause" text
            freegfx_begin(false);
            own_draw_box(13*HUNIT+NORM_OFFSET, 19, 32 , 11, DM_CLEAR);
            own_draw_rect(13*HUNIT+NORM_OFFSET, 19, 13*HUNIT+NORM_OFFSET+31, 19+10, 1, DM_SET);
            own_print_string((const nokstr_t*)"pause", 13*HUNIT+NORM_OFFSET+2, 20+6, app.font_bold, DM_SET);
            freegfx_end(false);

            //print current shape
            //TODO: blinking ;)
            tetris_print_shape ( game->shape_no, game->xpos + NORM_OFFSET, game->ypos, game->rot, 1 );
            #if defined DEBUG
               printf("Game paused\n");
            #endif
         }
      }
   }
}

/*
//file system
void load_settings()
{
   #if defined TFS
      if(TFS_SIZE(tfsstat(filename))!=sizeof(struct settings))
      {
   #endif
         if(settings)
            free(settings);
         settings=malloc(sizeof(struct settings));
         memset(game,0,sizeof(struct settings));

         //default settings
         settings->total_shapes=7;
         settings->stats=0;
         settings->hint=1;
         settings->s96=(screen_info.width==96);
         settings->blockdesign = 0;
         settings->level=1;
         settings->top_score=0;
   #if defined TFS
         //save to file
         tfserror(tfsadd(filename, TFS_A_HIDDEN, settings, sizeof(struct settings)));
      }
      else
      {
         if(settings)
            free(settings);
         settings=tfsload(filename);
      }
   #endif
}

void save_settings(void)
{
   #if defined TFS
      tfserror(tfsadd(filename, TFS_A_HIDDEN, settings, sizeof(struct settings)));
   #endif
}
*/
void play_sound(const ringtone_t *tone)
{
   if(own_get_game_setting(GS_SOUNDS))
      own_play_tone(tone,PTM_WARNING_TONE);
}

static void tetris_start_timer(void)
{
   own_timer_start(&timer,TT_MESSAGE,(5-(settings->level))*5);
}

static void tetris_stop_timer(void)
{
   own_timer_abort(&timer);
}

