/*****************************************************************************************
 *
 *  Copyright 2006+ by Cieszak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ****************************************************************************************/

#ifndef GAME_H
#define GAME_H

#include <nokix/all.h>

/**************************************************************************************/

//if You want tfs file system (default)
//#define TFS

//#define DEBUG


#define UWIDTH          10
#define UHEIGHT         15

// settings->s96 replaced with compile-time selection, allows optimizations by compiler
#ifndef S96
#define S96   0   	// default to 84x48 unless S96 already defined
#endif

#define VUNIT           (3+S96)              /* ver unit */
#define HUNIT		      (3+S96)              /* hor unit */
#define NORM_OFFSET	   1                    /* x offset for blocks */
#define NEXT_HEIGHT	   (VUNIT*4)            /* Height for next block */


struct game                      // structure sorted to pack bytes more efficiently (180 -> 172 B)
{
   unsigned initialized:1;       //if the game already initialized
   unsigned gameover:1;          //if gameover
   unsigned paused:1;            //if paused
   unsigned lock_keys:1;         //keys are locked (dropping shape, etc)

   unsigned shape_no:4;          //the dropping shape
   unsigned rot:2;               //rotation of shape_no
   unsigned next_no:4;           //next shape
   unsigned next_rot:2;          //rotation of next shape
   unsigned dlyloop:3;
   signed char xpos, ypos;       //x, y,  of shape_no
   unsigned short score;           //current score          // short should be enough
   unsigned short rows;            //number of rows deleted // short should be enough
   unsigned char grid[UWIDTH][UHEIGHT];
   unsigned char stats[10];      //shapes statistics
   unsigned char key;            //current pressed key
   unsigned char keycounter;
};

//settings_t moved to menu.h, it is now needed in struct application
//shape_t moved to game.c, because it does not need to be public
extern struct game *game;

/**************************************************************************************/

extern void free_game(void);
/*void tetris_define_shapes ( void );
void tetris_init_game ( void );
void tetris_create_shape ( void );
void tetris_show_score ( void );
void tetris_show_next ( void );
void tetris_print_shape ( int shape_no, int x, int y, int rot, int col );
void tetris_set1pixel ( int x, int y, int col );
void tetris_left_proc ( void );
void tetris_right_proc ( void );
void tetris_anti_proc ( void );
void tetris_fast_proc ( void );
int tetris_block_can_drop ( int shape_no, int xpos, int ypos, int rot );
int tetris_block_can_left ( int shape_no, int xpos, int ypos, int rot );
int tetris_block_can_right ( int shape_no, int xpos, int ypos, int rot );
void tetris_remove_full_lines ( int y );
int tetris_check_rot ( int shape_no, int xpos, int ypos, int newrot );
void tetris_print_all_blocks ( void );
void tetris_drop_block ( void );
void tetris_end_game ( void );
void tetris_create_shape ( void );
void tetris_store_shape ( int shape_no, int xpos, int ypos, int rot );
int tetris_level ( void );
void tetris_pause(int pause);
void tetris_print_board(void);
void tetris_print_stats(void);
void tetris_print_hint(int x, int t);

void load_settings(void);
void save_settings(void);*/

extern void play_sound(const ringtone_t *tone);

/*void tetris_start_timer(void);
void tetris_stop_timer(void);*/

/**************************************************************************************/

#endif /* GAME_H */

