/*****************************************************************************************
 *
 *  Copyright 2006+ by Cieszak
 *  Modified and optimized by NuukiaWorld
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ****************************************************************************************/

#include "menu.h"
#include "game.h"

#define OMSG_START_GAME     	(0x1f00) // message sent by our "New game" and "Continue" menu items (see _project.nrx)
#define OMSG_INSTRUCTIONS   	(0x1f01) // message sent by our "Instructions" menu item (see _project.nrx)
#define OMSG_OPTIONS       	(0x1f02) // option changed
#define OMSG_TOP_SCORE      	(0x1f03)
#define OMSG_CLEAR_TOP_SCORE  (0x1f04)

/**************************************************************************************/

extern rexx_uint game_dispatcher_id;              // dispatcher id of the game

extern const struct menu_item i_continue;           // menu items
extern const struct menu_item i_new_game;
extern const struct menu_item i_instructions;
extern const struct menu_item i_options;
extern const struct menu_item i_top_score;
extern const struct menu_item i_clear_top_score;

/*extern const struct menu_item i_kind;	// options menu is now a static menu
extern const struct menu_item i_level;
extern const struct menu_item i_stat;
extern const struct menu_item i_hint;
extern const struct menu_item i_block;

//stat and hint options
extern const struct menu_item i_staton;
extern const struct menu_item i_statoff;
extern const struct menu_item i_hinton;
extern const struct menu_item i_hintoff;
//kind optiond
extern const struct menu_item i_7;
extern const struct menu_item i_10;
//block design options
extern const struct menu_item i_1;
extern const struct menu_item i_2;
extern const struct menu_item i_3;*/

extern const struct menu menu_level;	// to trap the level menu access

extern const nokstr_t str_game_name;
extern const nokstr_t str_instr_text;
//extern const nokstr_t str_options;
extern const nokstr_t str_top_score_cleared;

// Non-volatile storage ids
extern rexx_uint confid_stats;
extern rexx_uint confid_hint;
extern rexx_uint confid_blockdesign;
extern rexx_uint confid_total_shapes;
extern rexx_uint confid_level;
extern rexx_uint eprid_top_score;

static const struct level_dialog level_dialog={NULL,0,NULL,1,4};

static const ULONG tim_disp_msg=MSG_SHOW_PENDING_DISP_MSGS;

/**************************************************************************************/

static void make_game_menu(void)
{
   while(menu_get_item(1,1)) menu_rem_item(1,1);

   if(game)
   {
//      if(game->gameover)	// game is freed in MSG_D_QUIT_NOW instead
//         free_game();
//      else
         menu_add_item(1,0,&i_continue);
   }
   menu_add_item(1,0,&i_new_game);
   menu_add_item(1,0,&i_options);
   menu_add_item(1,0,&i_top_score);
   menu_add_item(1,0,&i_clear_top_score);	// I like to show this always, save also some bytes ;)
   menu_add_item(1,0,&i_instructions);		// I like to have this last
}

/*static void make_options_menu(void)
{
   menu_delete(2);
   menu_create(2,MDM_LIST,MCM_DYNAMIC_CONFID|MCM_MARK_ID,1,&str_options);
   menu_add_item(2,0,&i_kind);
   menu_add_item(2,0,&i_stat);
   menu_add_item(2,0,&i_hint);
   menu_add_item(2,0,&i_block);
   menu_add_item(2,0,&i_level);
}

static void make_options_submenu(unsigned char arg)
{
   menu_delete(3);
   menu_create(3,MDM_LIST,MCM_DYNAMIC_CONFID|MCM_MARK_ID,1,&str_options);
   if(arg==1) //kind
   {
      menu_add_item(3,0,&i_7);
      menu_add_item(3,0,&i_10);
   }
   else
   if(arg==2)//stat
   {
      menu_add_item(3,0,&i_staton);
      menu_add_item(3,0,&i_statoff);
   }
   else
   if(arg==3)//hint
   {
      menu_add_item(3,0,&i_hinton);
      menu_add_item(3,0,&i_hintoff);
   }
   else
   if(arg==4)//block design
   {
      menu_add_item(3,0,&i_1);
      menu_add_item(3,0,&i_2);
      menu_add_item(3,0,&i_3);
   }
   else own_play_tone(&tone_game_lost, PTM_WARNING_TONE);
}*/

/**************************************************************************************/

/* UI messages dispatcher */
extern rexx_uint menu_dispatcher_id; /* dispatcher's call id */
rexx("menu_dispatcher_id=add_5e0_id(menu_dispatcher)"); /* auto creation of the id */

unsigned menu_dispatcher(unsigned msg)
{
   switch(msg)
   {
      case MSG_D_INIT: // initialize and open menu
         // Load those settings that are not changed by the menu directly
         app.font_plain = own_get_font("small/plain");
         app.font_bold = own_get_font("small/bold");
         settings->level = own_get_config(confid_level) + 1;
         eeprom_read( &settings->top_score, eprid_top_score, NULL);
         
         menu_create(1,0x60,MCM_MARK_ID,1,&str_game_name);
         make_game_menu();
         menu_open(1,1);
         app.state=STATE_MENU;
         msg=MSG_R_STAY;
         break;

      case MSG_D_QUIT: // handle menu quit
         menu_delete(1);
//         if(game)	// game is freed in MSG_D_QUIT_NOW
//            if(game->gameover)
//               free_game();
         msg=MSG_R_QUIT;
         break;

// Changed case order to allow fall-through to MSG_D_QUIT_NOW if malloc() fails
      case OMSG_START_GAME: // "New game" or "Continue" was selected from menu
         if(!game)
         {
            game=malloc_try(sizeof(struct game));	// use malloc_try to avoid hang if memory is full
         }

         if( game != NULL ) {
            /* since OMSG_START_GAME message is sent by both "New game" and "Continue"
               items, we have to distinguish between them and initialize the game to
               start values only when "New game" was selected;
               we do it by checking the argv[0] of the message (it was set in _project.nrx) */
            if(message.argv[0])
            {
               // TODO: game continue handling
            }
            else
            {
               //new game
               memset(game,0,sizeof(struct game));
               //game->initialized=0;
            }

            // Load those settings that can be changed directly by the menu
            settings->stats = own_get_config(confid_stats);
            settings->hint = own_get_config(confid_hint);
            settings->blockdesign = own_get_config(confid_blockdesign);
            settings->total_shapes = ( own_get_config(confid_total_shapes) ? 10 : 7 );

            // start the game dispatcher
            send_message(MSG_START_HANDLER|MSG_1P,game_dispatcher_id); /* start the game */

            app.state=STATE_GAME;

            msg=MSG_NOP;
            break;
         }
         // FALL-THROUGH - out of memory - reopen menu
      case MSG_D_QUIT_NOW: // called when we get back from some other dispatcher
      {
//         unsigned char oldstate=app.state;	// not needed, STATE_MENU is set at the end of case now

         switch(app.state) {
            case STATE_LEVEL:
               if(message.argv[0]==QNS_OK)
               {
                  own_set_config( confid_level, settings->level - 1);
                  free_game();
                  display_msg(DMI_DONE, &str_done);
               }
               break;
            case STATE_GAME:
               if(game->gameover) {
               		// top score message moved to here, then there is no need to access any nv storage in game.c
                  if(game->score>settings->top_score)
                  {
                     settings->top_score=game->score;
                     eeprom_write( &settings->top_score, eprid_top_score, NULL );
                     play_sound(&tone_game_won);
               //      send_message(MSG_DISPLAY_MESSAGE|MSG_2P,DMI_INFORMATION,w_sprintf("Najleprzy: %d",game->score));
                     send_message(MSG_DISPLAY_MESSAGE|MSG_2P,DMI_CUSTOM,w_sprintf(&str_game_over_top_score_n,game->score));
                     own_timer_start(&tim_disp_msg,TT_MESSAGE,3*TICKS_PER_SECOND);
//                     app.state = STATE_MESSAGE;
                  }
                  free_game();
               }
               break;
         }
         app.state=STATE_MENU;
         make_game_menu();
         menu_reopen();
         msg=MSG_R_STAY;
         break;
      }

      case OMSG_INSTRUCTIONS: // "Instructions" was selected from menu
         // displays the instructions screen
         send_message(MSG_START_HANDLER|MSG_2P,id_game_instructions,str_instr_text);

         app.state=STATE_INSTRUCTIONS;

         msg=MSG_NOP;
         break;

      case OMSG_TOP_SCORE:
         send_message(MSG_DISPLAY_MESSAGE|MSG_2P,DMI_TOP_SCORE,w_sprintf(&str_top_score_n,settings->top_score));
         menu_reopen();
         msg=MSG_NOP;
         break;

// clear top score has now its own msg
      case OMSG_CLEAR_TOP_SCORE:
         dialog_open((char *)&str_are_you_sure,&dialog_bool);
         app.state=STATE_CLEAR_DLG;
         msg=MSG_NOP;
         break;

      case MSG_DIALOG_ACCEPTED:
         if(app.state==STATE_CLEAR_DLG)
         {
            //clearing top score
            settings->top_score=0;
            eeprom_write( &settings->top_score, eprid_top_score, NULL );
            send_message(MSG_DISPLAY_MESSAGE|MSG_2P,DMI_INFORMATION,&str_top_score_cleared);
            make_game_menu();
            menu_reopen();
            app.state=STATE_MENU;
         }
         break;

/*      case OMSG_OPTIONS:
         switch(message.argv[0])
         {
            case 0:
               app.state=STATE_OPTIONS_MENU;
               make_options_menu();
               menu_open(2,1);
               msg=MSG_NOP;
               break;
               
            case 1: //kind
               app.state=STATE_OPTIONS_SUBM;
               make_options_submenu(message.argv[0]);
               menu_open(3, (settings->total_shapes==7?1:2));
               msg=MSG_NOP;
               break;

            case 2: //stat
               app.state=STATE_OPTIONS_SUBM;
               make_options_submenu(message.argv[0]);
               menu_open(3, (settings->stats?1:2));
               msg=MSG_NOP;
               break;

            case 3: //hint
               app.state=STATE_OPTIONS_SUBM;
               make_options_submenu(message.argv[0]);
               menu_open(3, (settings->hint?1:2));
               msg=MSG_NOP;
               break;

            case 4: //block design
               app.state=STATE_OPTIONS_SUBM;
               make_options_submenu(message.argv[0]);
               menu_open(3, settings->blockdesign+1);
               msg=MSG_NOP;
               break;

            case 5: //level
               send_message(MSG_START_HANDLER|MSG_3P,id_level_dialog,&(settings->level),&level_dialog); // open level dialog
               app.state=STATE_LEVEL;
               msg=MSG_NOP;
               break;

         //sub options
            case 11: //7 shapes
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->total_shapes=7;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 1);
               msg=MSG_NOP;
               break;

            case 12: //10 shapes
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->total_shapes=10;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 1);
               msg=MSG_NOP;
               break;

            case 21: //stats on
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->stats=1;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 2);
               msg=MSG_NOP;
               break;

            case 22: //stats off
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->stats=0;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 2);
               msg=MSG_NOP;
               break;

            case 31: //hint on
               menu_close();
               app.state=STATE_OPTIONS_MENU;
               settings->hint=1;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 3);
               msg=MSG_NOP;
               break;

            case 32: //hint off
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->hint=0;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 3);
               msg=MSG_NOP;
               break;

            case 41: //first block design
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->blockdesign=0;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 4);
               msg=MSG_NOP;
               break;

            case 42: //second block design
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->blockdesign=1;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 4);
               msg=MSG_NOP;
               break;

            case 43: //third block design
               app.state=STATE_OPTIONS_MENU;
               menu_close();
               settings->blockdesign=2;
               save_settings();
               display_msg(DMI_DONE, &str_ok);
               menu_open(2, 4);
               msg=MSG_NOP;
               break;
         }
         break;*/

      case MSG_BACK:	// only back in clear top score dialog needs to be processed
         if( app.state == STATE_CLEAR_DLG ) {
            dialog_close();
            app.state = STATE_MENU;
            menu_reopen();
            msg=MSG_NOP;
         }
         break;

// Trap the message opening the level menu, show a level dialog instead
      case MSG_OPEN_STATIC_MENU:
         // Show level dialog instead of level menu!
         // The menu is defined only to show the selected level in the options menu
         if( msg_argv[0] == (unsigned int)&menu_level ) {
            send_message(MSG_START_HANDLER|MSG_3P,id_level_dialog,&(settings->level),&level_dialog); /* open level dialog */
            app.state=STATE_LEVEL;
            msg=MSG_NOP;
         }
         break;

/*      case OMSG_OPTION_CHANGE: // delete current game if any option is changed, unnecessary
         free_game();
         msg=MSG_NOP;
         break;*/
   }
   return msg;
}

/**************************************************************************************/

